## ############################################################################
## correlate - merge the photometry of multiple sources from multiple 
##             bands by correlating their magnitudes and errors based 
##             on their WCS coordinates
##
## (c) 2005 SAO, Mike Schuster - mschuster@cfa.harvard.edu
##
## ############################################################################

=head1 NAME

 perl correlate.pl - merge the photometry of multiple sources from multiple 
                     bands by correlating their magnitudes and errors based 
                     on their WCS coordinates


=head1 SYNOPSIS

 % perl correlate.pl [--help][--ver][--t tolerance][--N max_iter][--r radius]
                     [--b boxscale][--a][--e max_error][--s min_S2N]
                     [--c red_color_cut][--phot phot_file_root]
                     [--wcs wcs_file_root][--o output_file_root]


=head1 COMMAND LINE OPTIONS

 --help
   Print usage information and exit.

 --ver
   Print current version and exit.

 --t tolerance
   tolerance in arcseconds for maximum average differential in arcseconds 
   (Dec or projected RA) between channels between successive iterations 
   - default is 0.01

 --N max_iter
   maximum number of iterations to achieve tolerance 
   - default is 3

 --r radius
   maximum initial separation in arcseconds for matching sources 
   - default is 2.444

 --b boxscale
   search box size scale for matching sources
   positive - in multiples of matching radius
   negative - in integral arcminutes
   - default is 4

 --a
   use progressively more aggressive initial separation constraint with each 
   successive iteration 
   - not used as default

 --e max_error
   maximum photometric error allowed (in magnitudes) 
   - default is 1.0857 (i.e. S/N=1)

 --s min_S2N
   minimum signal-to-noise allowed 
   - default is 1 (i.e. max_error=1.0857)

 --c red_color_cut
   red source color cut (in magnitudes): ch1-ch2 >= red_color_cut
   - default is 1.5

 --phot phot_file_root
   Photometry file: phot_file_root.channel 
   - default is "magd"

 --wcs wcs_file_root
   WCS coordinate file: wcs_file_root.channel 
   - default is "wcscood"

 --o output_file_root
   Output correlated photometry file root
   - default is "phot"


=head1 DESCRIPTION

This routine merges the photometry of multiple sources from up to 4 bands by 
correlating their magnitudes and errors based on their WCS coordinates. The 
routine compares each source with the neighboring sources in the other bands 
to find the closest matching source for all of the other bands. First, 
sources in band 1 are compared to those in bands 2, 3, and 4. Second, 
unmatched sources in band 2 are compared to unmatched sources in bands 3 
and 4. Finally, the remaining unmatched sources in band 3 are compared to 
the remaining unmatched sources in band 4. This cascading matching process 
is repeated until any one of the exit criteria described below are satisfied.

The output files are (-o option) phot.all and phot.red (sources with 
m[3.6]-m[4.5] >= red_color_cut, provided those bands are present), SAOImage 
DS9 region files phot.all.reg and phot.red.reg, and a log file phot.log. The 
correlated photometry files phot.all and phot.red have the format:

source hh mm ss. dd mm ss. mag1 err1 mag2 err2 mag3 err3 mag4 err4

The WCS coordinates are those of the primary band, that is, the coordinates 
for band 1 and its matching sources are the band 1 coordinates. In the DS9 
region files the radius of the circle is equal to the matching radius.

The input photometry files (magd.?, -phot option) that contain the magnitudes 
and errors should have the format: mag err, one pair per line (err can also 
be blank or INDEF). The input WCS coordinate files (wcscood.?, -wcs option) 
should contain the sexagesimal RA and Dec in the SAME order AND number as in 
the input photometry files (each channel can have a different number of 
sources) with the format: hh:mm:ss. dd:mm:ss., one pair of coordinates per 
line. The IRAF task "pdump" will extract the magnitudes, errors, and 
coordinates of sources in an IRAF photometry file based on several useful 
expressions. The IRAF task "wcsctran" can be used to convert the extracted 
coordinate lists into sexagesimal WCS coordinates. These lists need not be 
ordered in any particular way, as long as the photometry and WCS coordinate 
lists match. That is to say, the sources can be randomly positioned on the 
sky.

As the input magnitude and coordinate lists are read they are trimmed of 
low-quality data based on the setting of either the -e (max error) or the -s 
(min S/N) option (but not both). At this point the routine produces a copy of 
the trimmed input photometry file (magd_phot.?) and the accompanying WCS list 
(wcscood_phot.?) with their overall ordering unchanged. Next, the sexagesimal 
WCS coordinates are shifted by the average correction determined from the 
previous iteration (except the primary band) and converted to decimal 
degrees. Each source is placed in a search box based on its WCS coordinates 
to expedite the matching process. These search boxes can be adjusted with the 
-b option (search box size scale - positive for multiples of the matching 
radius, negative for integral arcminutes). The RA bins are 4 to 6.28x larger 
than the Dec bins to minimize binning shear as a function of Dec. RA bin 
shear is kept to a maximum of 1 box. There are no problems with the code at 
the poles, or across the ecliptic or 0/24 hr meridian.

Source matching is done by comparing each primary source in a given box with 
the secondary sources in the overlapping box, as well as the surrounding 
boxes in the secondary bands (to catch possible matches across box 
boundaries), checking that the box exists for the secondary bands. 
Secondary sources are considered a match when their corrected decimal WCS 
coordinates are less than the matching radius (-r option) from the primary 
source. The matching process can be made more aggressive with the -a option, 
where the matching radius is reduced by 2/3 on the second iteration, and by 
1/2 on the third and subsequent iterations. Candidate matches are compared 
against one another to find the best possible match for each primary source. 
Orphaned primary sources, those sources in the primary band that had the 
best match taken away by another primary source later in the list, are 
compared with all remaining unmatched secondary sources in the overlapping 
and surrounding boxes in the secondary channel, again comparing candidate 
matches to find the best available.

The final ordering of the sources in the output photometry files depends on 
the absolute size of the search box. This is because the routine assumes no a 
priori ordering of the input data. Within each search box primary sources are 
listed according to their original ordering. Smaller search boxes will result 
in a finer ordering of the output data (see warnings below). Overall, the 
output is ordered roughly in RA (and then Dec) since the search boxes are 
considered in this order.

The routine will repeat the matching process until it has reached: (1) the 
maximum number of iterations specified with the -N option, or (2) when the 
maximum absolute value of the average offset shift for the previous iteration 
is less than the desired tolerance specified by the -t option.

The offset shift statistics are output to the terminal and to the log file 
phot.log with the format:

Register WCS to channel 1 (shifts are in arcseconds)
Exit after step n:
 
ch  RA_shift Dec_shift  dRA dDec  |dRA| |dDec|  rms_dRA rms_dDec  matches

where RA_shift and Dec_shift are the total shifts applied to that band's (ch 
- channel's) data through the last iteration. dRA and dDec are the average 
offset shifts for the last iteration (these are compared to the tolerance in 
absolute value). |dRA| and |dDec| are the averages of the absolute offsets, 
and rms_dRA and rms_dDec are the usual root mean square differentials (these 
are the half-widths of the radial-offset and average-offset distributions, 
respectively). The final number of matches between the primary and secondary 
channels is listed in the last column. The average offset shift is calculated 
between matched sources for bands 1 & 2, bands 1 & 3 and bands 1 & 4. If band 
1 is not present, then the average offset shift is calculated between matched 
sources for bands 2 & 3 and bands 2 & 4, and so on if band 2 is not present.

shift error = sqrt(rms**2 - d**2) / sqrt(matches)
            = rms deviation / sqrt(matches), when d=0

|d| error = sqrt(rms**2 - |d|**2) / sqrt(matches)

=head2 Hints:

Try to set the matching radius >= 3*rms deviation (from output). Smaller than 
this and your probably missing real sources. If the matching radius is too 
large, then the routine may erroneously match sources and the rms deviation 
will likely grow larger.

There seems to be a sweet spot with IRAC data for the absolute size of the 
search box scale, in terms of speed, that is around 10 arcseconds.


=head2 Special values for options:

"0" is interpolated as null, or "" (blank), and will result in default values
To set an option to zero use "0."
Setting tolerance (-t), max_iter (-N), radius (-r), or boxsize (-b) to "0." 
will cause the routine to display options and exit.

=head2 Warnings:

Processing time and memory conservation means that most correlation scripts, 
including this one, are not perfect. This routine does a reasonable job at 
finding the best possible match for each primary source. With this routine a 
change in the search box size will alter the final ordering of the data (no 
matter the initial ordering). As a consequence of the matching algorithm 
employed in this routine, any change in the ordering of the data may result 
in a statistically insignificant change in the number of matched sources. In 
addition, the specific matches between a statistically insignificant number 
of sources may also change.

Setting the absolute size of the search box too small will cause the routine 
to use too much memory and greatly reduce speed. In this case, more sources 
in a box is better than more search boxes. Conversely, setting the absolute 
size of the search box too large will also cause the routine to run slowly 
(roughly a doubling in time, from the minimum, with a doubling in box size, 
up to a size comparable to the boundaries for the data). The search is a 
correlation (and so, goes as the number of search boxes and sources squared) 
and does not assume the input is ordered in any particular way (so it runs in 
memory), and so care must be taken to do the source matching efficiently. 
There is a built-in check on the number of search boxes (i.e. absolute size 
of the search box specific to your data set).

=head2 Limitations:

It is assumed that small angle approximations are valid for the offset shifts 
and calculating the angular separation between sources. However, the actual 
angular separation between sources is correctly given by:

# $cp = cos(3.14159*($dec[$f][$k])/180.);
# $cs = cos(3.14159*($dec[$fn][$kn])/180.);
# $sp = sin(3.14159*($dec[$f][$k])/180.);
# $ss = sin(3.14159*($dec[$fn][$kn])/180.);
# $x = cos(3.14159*($ra[$f][$k]-$ra[$fn][$kn])/180.)*$cp*$cs - $sp*$ss;
# $diff = 3600.*180.*atan2(sqrt(1-$x**2),$x)/3.14159;

But, it is much simpler to use:

# $ddec = ($dec[$f][$k]-$dec[$fn][$kn])*3600.;
# $dra = ($ra[$f][$k]-$ra[$fn][$kn])*3600.*cos(3.14159*$dec[$f][$k]/180.);
# $diff = ($dra**2+$ddec**2)**0.5;

This should be valid for all imaginable applications of this routine since 
the matching radius should be small enough that small angle approximations 
are valid. If greater precision is required, then replace the appropriate 
sections of code. There is no need to alter the offset shift correction 
between bands since the routine finds a solution iteratively.


=head1 CHANGES

 1.4.3 - beta version
         - updated documentation
         (Sept 15, 2006)
 1.4.2 - beta version
         - updated documentation
         (May 30, 2006)
 1.4.1 - beta version
         - fixed bug when RA coords are >=720 or <0
         (Mar 29, 2006)
 1.4 - beta version
       - routine now works across 0/24 hr meridian
       - added option to set red source color cut (for phot.red output)
       (Mar 2, 2006)
 1.3 - beta version
       - fixed bug with determining box for RA coords 
       (Jan 18, 2006)
 1.2 - beta version
       - will now run on any combination of channels, up to 4
       - fixed problem with determining box for RA coords 
       (Jan 13, 2006)
 1.1 - beta version (correlate.pl) (Jan 06, 2006)
 1.0 - early version of correlate.pl (Dec 22, 2006)
 0.x - several previous versions (getphot.pl) by several authors, SAO


=head1 AUTHOR

 (c) 2005 Mike Schuster (mschuster@cfa.harvard.edu)
 (c) 2005 SAO - Smithsonian Astrophysical Observatory


=head1 SEE ALSO



=head1 BUGS

Report bugs to Mike Schuster, mschuster@cfa.harvard.edu


=cut

## ############################################################################
## Set Modules and version number
## ############################################################################

#system "date";

use warnings;
use strict;
use diagnostics;
use Getopt::Long;

# Set number version
my $VERSION = '1.4.3 - beta';           # Script version number
my $pname   = 'correlate.pl';    # Script name


## ############################################################################
## Read input options
## ############################################################################

my ($help,$ver);
my ($toler,$iter,$rad,$boxscale,$a,$max_err,$min_s2n,$color);
my ($phot,$wcs,$output);

my %optctl = (help            => \$help,
              ver             => \$ver,
              t               => \$toler,
              N               => \$iter,
              r               => \$rad,
              b               => \$boxscale,
              a               => \$a,
              e               => \$max_err,
              s               => \$min_s2n,
              c               => \$color,
              phot            => \$phot,
              wcs             => \$wcs,
              o               => \$output,
             );

&GetOptions(\%optctl,
            "help",
            "ver",
            "t=f",
            "N=f",
            "r=f",
            "b=f",
            "a",
            "e=f",
            "s=f",
            "c=f",
            "phot=s",
            "wcs=s",
            "o=s",
           );


## ############################################################################
## Process input options
## ############################################################################

# Print help and exit
my $usg = "USAGE:  perl $pname [--help][--ver][--t tolerance][--N max_iter]\n"
         ."                            [--r radius][--b boxscale][--a]\n"
         ."                            [--e max_error][--s min_S2N]\n"
         ."                            [--c red_color_cut]\n"
         ."                            [--phot phot_file_root]\n"
         ."                            [--wcs wcs_file_root]\n"
         ."                            [--o output_file_root]\n";

if ($help) {
  print STDOUT <<HELP;

  $pname version $VERSION

  Merge the photometry of multiple sources from multiple bands by 
  correlating their magnitudes and errors based on their WCS coordinates

  $usg

HELP
  exit 0;
}


# Print version and exit
if ($ver) { 
  print "$pname: version $VERSION\n"; 
  exit 0;
}

# Check the options
if ($max_err and $min_s2n) {
 print "$pname: ERROR: cannot set both max photometric error and min S/N\n";
 exit 0;
}

if (not $toler) { $toler = 0.01; }
else {
 if ($toler<0.) {
  print "$pname: WARNING: tolerance cannot be negative\n";
  print "$pname: WARNING: setting tolerance to default\n";
  $toler = 0.01;
 }
}

if (not $iter) { $iter = 3; }
else {
 if ($iter<0.) {
  print "$pname: WARNING: number of iterations cannot be negative\n";
  print "$pname: WARNING: setting number of iterations to 1\n";
  $iter = 1;
 }
 $iter = int($iter);
}

if (not $rad) { $rad = 2.444; }
else {
 if ($rad<0.) {
  print "$pname: WARNING: matching radius cannot be negative\n";
  print "$pname: WARNING: setting matching radius to default\n";
  $rad = 2.444;
 }
}

my ($box,$bsam);
if (not $boxscale) { $boxscale = 4; }
else {
 if (($boxscale<1.1) and ($boxscale>0.)) {
  print "$pname: WARNING: search box must be larger than matching radius\n";
  print "$pname: WARNING: setting search box size scale to 1.1 units\n";
  $boxscale = 1.1;
 }
 if (($boxscale<0.) and ($boxscale > -1.)) {
  print "$pname: WARNING: (negative -b) search box size scale must be at least 1\'\n";
  print "$pname: WARNING: setting search box size scale to 1\'\n";
  $boxscale = -1;
 }
}
if (($boxscale <= -1) and ($rad != 0.)) {
 $bsam = int(abs($boxscale));
 $boxscale = $bsam*60./$rad;
 if ($boxscale<1.1) {
  print "$pname: WARNING: search box must be larger than matching radius\n";
  print "$pname: WARNING: setting search box size scale to 1.1 units\n";
  $boxscale = 1.1;
  undef($bsam);
 }
}
$box=$boxscale*$rad;

my $aggressive = 0;
my $a_label = "no";
if ($a) { $aggressive = 1; $a_label = "yes"; }

if ((not $max_err) and (not $min_s2n)) { $max_err = 1.0857; $min_s2n = 1.0857/$max_err; }
if ($max_err) {
 if ($max_err<=0.) {
  print "$pname: WARNING: maximum photometric error must be positive\n";
  print "$pname: WARNING: setting maximum photometric error to default\n";
  $max_err = 1.0857;
 }
 $min_s2n = 1.0857/$max_err;
}

if ($min_s2n) {
 if ($min_s2n<0.) {
  print "$pname: WARNING: minimum signal-to-noise cannot be negative\n";
  print "$pname: WARNING: setting minimum signal-to-noise to default\n";
  $min_s2n = 1.
 }
 if ($min_s2n==0.) {
  $max_err = "INDEF"; }
 else {
  $max_err = 1.0857/$min_s2n; }
}

if (not $color) { $color = 1.5; }

if (not $phot) { $phot = "magd"; }
if (not $wcs) { $wcs = "wcscood"; }
if (not $output) { $output = "phot"; }

if ($rad<$toler) {
 print "$pname: ERROR: must set tolerance <= matching radius\n";
 exit 0;
}

if ($min_s2n != 0.) {
 $max_err = sprintf "%.4f",$max_err; }
$min_s2n = sprintf "%.2f",$min_s2n;

# Print settings
print "\n";
print "USAGE:  perl $pname [--help][--ver]\n";
print "VERSION:  $VERSION\n";
print "\n";
print "Tolerance for average dRA and dDec = ",$toler,"\"\n";
print "Maximum number of iterations = ",$iter,"\n";
print "Matching radius = ",$rad,"\"\n";
if ($bsam) {
 print "Search box size scale = ",$bsam,"\'\n"; }
else {
 print "Search box size scale = ",$boxscale," * matching radius\n"; }
print "Aggressive Mode?: ",$a_label,"\n";
print "Maximum photometric error = ",$max_err," mag.\n";
print "Minimum signal-to-noise = ",$min_s2n,"\n";
print "Red source color cut: ch1-ch2 >= ",$color,"\n";
print "Photometry file root: ",$phot,"\n";
print "WCS coordinate file root: ",$wcs,"\n";
print "Correlated photometry file root: ",$output,"\n";
print "\n";

if (($iter==0.) or ($toler==0.) or ($rad==0.) or ($boxscale==0.)) {
 print "$pname: display options and exit\n";
 exit 0;
}


## ############################################################################
## Read in photometry and WCS coordinate files, make copies of trimmed lists
## ############################################################################

my ($c,$f,$h,$i,$j,$k,$m,$itot,$maxf);
my (@mag,@sig,@num);
my (@ra,@dec,@hr,@min,@sec,@deg,@am,@as);
my (@cf,@fc,@fi,@kfi);
my (@dx0,@dy0);

# set values for blank entries
for ($c=2; $c<+5;$c++) {
    $cf[$c]="";
    $cf[$c]="";
    $fc[$c]="";
    $fc[$c]="";
}

$f=0;
for ($c=1; $c<+5;$c++) {

    my @mag2wcs;
    my ($num_mag,$num_wcs,$num_inmag,$num_inwcs);
    my $file1=join '',($phot,".",$c);
    my $file2=join '',($wcs,".",$c);
    my $file1c=join '',($phot,"_",$output,".",$c);
    my $file2c=join '',($wcs,"_",$output,".",$c);

    if ((-f $file1) and (-f $file2)) {
    $f=$f+1;
    $cf[$f]=$c;
    $fc[$c]=$f;

     # Read in magnitude file and trim unwanted values
     $i=1;
     $j=0;
     open (FILE1c,"> $file1c") or die "$pname: cannot open file to write to\n";
     open (FILE1,"$file1") or die "$pname: cannot open photometry file ",$file1,"\n"; 
     while (<FILE1>) {
      next if /^$/;           # skip blank lines
      next if /^#/;           # skip commented lines
      my ($mag_i,$sig_i)=split;
      $j=$j+1;
      if ($mag_i) {
       $mag[$f][$i]=$mag_i; }
      else {
       $mag[$f][$i]="na"; }
      if ($sig_i) {
       $sig[$f][$i]=$sig_i; }
      else {
       $sig[$f][$i]="na"; }
      if (($sig[$f][$i] ne "na") and ($sig[$f][$i] ne "INDEF")) {
       if ($sig[$f][$i]>=10) {
        $sig[$f][$i]=sprintf "%5.2f",$sig[$f][$i]; }}
      if ($min_s2n != 0.) {
       if (($mag[$f][$i] ne "INDEF") and ($sig[$f][$i] ne "INDEF") and ($mag[$f][$i] ne "na") and ($sig[$f][$i] ne "na") and ($sig[$f][$i] <= $max_err)) {
        printf FILE1c "%8.3f%8.3f\n",$mag[$f][$i],$sig[$f][$i];
        $mag2wcs[$j]=$i;
        $i=$i+1;
       }
      }
      else {
       if (($mag[$f][$i] ne "INDEF") and ($mag[$f][$i] ne "na")) {
        printf FILE1c "%8.3f%8s\n",$mag[$f][$i],$sig[$f][$i];
        $mag2wcs[$j]=$i;
        $i=$i+1;
       }
      }
     } #end while loop
     close FILE1;
     close FILE1c;
     $num_inmag=$j;
     $num_mag=$i;

     # Read in WCS coordinate file and trim values that match trimmed mags
     $i=1;
     $j=0;
     open (FILE2c,"> $file2c") or die "$pname: cannot open file to write to\n";
     open (FILE2,"$file2") or die "$pname: cannot open WCS coordinate file ",$file2,"\n"; 
     while (<FILE2>) {
      next if /^$/;           # skip blank lines
      next if /^#/;           # skip commented lines
      my ($ra_i,$dec_i)=split;
      $j=$j+1;

      if ($mag2wcs[$j]) {
       ($hr[$f][$i],$min[$f][$i],$sec[$f][$i])=split(/:/,$ra_i);
       ($deg[$f][$i],$am[$f][$i],$as[$f][$i])=split(/:/,$dec_i);
       printf FILE2c "%13s%13s\n",$ra_i,$dec_i;
       $i=$i+1;
      }
     } #end while loop
     close FILE2;
     close FILE2c;
     $num_inwcs=$j;
     $num_wcs=$i;

     # Check that number of mags matches number of wcs coords
     if ($num_inmag>$num_inwcs) {
      print "$pname: ERROR: There are more input magnitudes than input WCS coordinates for channel ",$cf[$f],"\n";
      exit 0;
     }
     if ($num_inmag<$num_inwcs) {
      print "$pname: ERROR: There are more input WCS coordinates than input magnitudes for channel ",$cf[$f],"\n";
      exit 0;
     }
     if ($num_mag>$num_wcs) {
      print "$pname: ERROR: There are more magnitudes than WCS coordinates for channel ",$cf[$f],"\n";
      exit 0;
     }
     if ($num_mag<$num_wcs) {
      print "$pname: ERROR: There are more WCS coordinates than magnitudes for channel ",$cf[$f],"\n";
      exit 0;
     }

     if ($num_mag == $num_wcs) {
      $num[$f] = $num_mag;
      if ($num[$f]>10000) {
       print "channel ",$cf[$f]," data set is large, it may take a while\n"; }
      if ($num[$f]==1) {
       print "channel ",$cf[$f]," data set is empty\n"; }
     }
    }
    else {
     if (not -f $file1) {
      print "channel ",$c,": No photometry file, will continue processing\n"; }
     if (not -f $file2) {
      print "channel ",$c,": No WCS coordinate file, will continue processing\n"; }
    }

undef(@mag2wcs);

} #end for loop

$maxf=$f+1;
if ($maxf==1) {
    print "$pname: ERROR: No data was read in\n";
    exit 0;
}


## ############################################################################
## Begin iterations
## ############################################################################

sub max_abs {
    my $max=shift(@_);
    my $val;
    foreach $val (@_) {
        $max=abs($val) if abs($max)<abs($val);
    }
    return abs($max);
}

my (@kb,@num_kb);
my (@min_rab,@max_rab,@min_decb,@max_decb);
my (@ns_min,@ns_max,@nd_min,@nd_max);

# initialize values
for ($f=1; $f<+$maxf;$f++) {
    $dx0[$f]=0.;
    $dy0[$f]=0.;
}

$h=1;
my $maxd=1.1*$toler;
my $scale=1;
my $maxb=int(90.*3600./$box);

# begin iterations
while (($h<$iter+1) and ($maxd>$toler)) {

my @zero;
undef(@zero);

my ($ns,$nd,$k_ra,$k_dec);
my ($rab,$decb);
my (@b);

undef(@b);

## #################################################
## Convert WCS coordinates to decimal and add shifts
## #################################################

# set initial f for coordinate conversion - only do first channel once
if ($h==1) {
    $m=1; }
else {
    $m=2; }

# unset and initialize values
for ($f=$m; $f<+$maxf;$f++) {
    $min_rab[$f][0]=int($maxb/2)+1;
    $max_rab[$f][0]=0;
    $min_rab[$f][1]=$maxb;
    $max_rab[$f][1]=0;
    $min_decb[$f]=$maxb;
    $max_decb[$f]=0;
    $ns_min[$f]=1;
    $ns_max[$f]=0;
    $nd_min[$f]=1;
    $nd_max[$f]=0;
}

for ($f=$m; $f<+$maxf;$f++) {
    for ($i=1; $i<+$num[$f];$i++) {

    # convert RA and Dec to decimal degrees
    if (($deg[$f][$i]<0) or ($deg[$f][$i] eq "-0") or ($deg[$f][$i] eq "-00")) {
     $dec[$f][$i]=$deg[$f][$i]-$am[$f][$i]/60.-($as[$f][$i]-$dy0[$f])/3600.; }
    else {
     $dec[$f][$i]=$deg[$f][$i]+$am[$f][$i]/60.+($as[$f][$i]+$dy0[$f])/3600.; }

    $ra[$f][$i]=0.;
    if (cos(3.14159*$dec[$f][$i]/180.) != 0.0) {
     $ra[$f][$i]=15.*($hr[$f][$i]+$min[$f][$i]/60.+($sec[$f][$i]+$dx0[$f]/(15.*abs(cos(3.14159*$dec[$f][$i]/180.))))/3600.); }
    else {
     $ra[$f][$i]=0.; }

    if (abs($dec[$f][$i])>90.) {
     if ($dec[$f][$i]>90.) {
      $dec[$f][$i]=180.-$dec[$f][$i];
      $ra[$f][$i]=$ra[$f][$i]+180.;
     }
     if ($dec[$f][$i]<-90.) {
      $dec[$f][$i]=-180.-$dec[$f][$i];
      $ra[$f][$i]=$ra[$f][$i]+180.;
     }
    }
    if (($ra[$f][$i]>=360.) or ($ra[$f][$i]<0.)) {
     if ($ra[$f][$i]>=360.) {
      $ra[$f][$i]=$ra[$f][$i]-360.*int($ra[$f][$i]/360.); }
     if ($ra[$f][$i]<0.) {
      $ra[$f][$i]=$ra[$f][$i]+360.*(int(abs($ra[$f][$i])/360.)+1); }
    }

    # set box for coords - will loop on this later
    # max number of possible RA boxes at ecliptic
    # must equal max number of Dec boxes possible
    # to keep RA binning shear to 1 box
    if ($dec[$f][$i]<0) {
     $ns=0;
     $ns_min[$f]=0;
     $decb=int(abs($dec[$f][$i])*3600./$box);
    }
    else {
     $ns=1;
     $ns_max[$f]=1;
     $decb=int($dec[$f][$i]*3600./$box);
    }

    if ($ra[$f][$i]<180) {
     $nd=0;
     $nd_min[$f]=0;
    }
    else {
     $nd=1;
     $nd_max[$f]=1;
    }
    $rab=int($ra[$f][$i]*($maxb-$decb+1)/360.);

    # set tag based on box
    if (not $b[$rab][$decb][$ns]) {
     $b[$rab][$decb][$ns]=1; }
    else {
     $b[$rab][$decb][$ns]=$b[$rab][$decb][$ns]+1; }
    $kb[$f][$rab][$decb][$ns][$b[$rab][$decb][$ns]]=$i;
    if ($rab>$max_rab[$f][$nd]) {
     $max_rab[$f][$nd]=$rab; }
    if ($rab<$min_rab[$f][$nd]) {
     $min_rab[$f][$nd]=$rab; }
    if ($decb>$max_decb[$f]) {
     $max_decb[$f]=$decb; }
    if ($decb<$min_decb[$f]) {
     $min_decb[$f]=$decb; }

    # exit if search parameters are too fine for data set - memory problems
    if (($max_rab[$f][0]-$min_rab[$f][0]>=4999) or ($max_rab[$f][1]-$min_rab[$f][1]>=4999) or ($max_decb[$f]-$min_decb[$f]>=4999)) {
     print "$pname: ERROR: The absolute size of the search box must be larger for this data set\n";
     exit 0;
    }

    } #end for loop i

    # set number of sources per channel in each box
    for ($nd=$nd_min[$f]; $nd<+$nd_max[$f]+1;$nd++) {
    for ($k_ra=$min_rab[$f][$nd]; $k_ra<+$max_rab[$f][$nd]+1;$k_ra++) {
    for ($k_dec=$min_decb[$f]; $k_dec<+$max_decb[$f]+1;$k_dec++) {
    for ($ns=$ns_min[$f]; $ns<+$ns_max[$f]+1;$ns++) {

     $num_kb[$f][$k_ra][$k_dec][$ns]=1;
     if ($b[$k_ra][$k_dec][$ns]) {
      $num_kb[$f][$k_ra][$k_dec][$ns]=$b[$k_ra][$k_dec][$ns]+1; }

    } #end for loop ns
    } #end for loop k_dec
    } #end for loop k_ra
    } #end for loop nd

undef(@b);

} #end for loop f


## #############################################
## Calculate correlated offsets between channels
## #############################################

# adjust search radius if aggressive mode
if ($aggressive) {
    if ($h<=3) {
     $rad=$rad*$scale;
     $scale=1+(($h-1)/2);
     $rad=$rad/$scale;
    }
}

my ($nsf,$ndf,$fn,$kn,$k_w,$kn_w,$kn_ra,$kn_dec,$fno,$io,$jtot);
my ($dra,$ddec,$diff);
my (@drao,@ddeco,@diffo,@match,@redo_f,@redo_i,@num_ch);
my (@dn,@dx,@dy,@dx_abs,@dy_abs,@dx_rms2,@dy_rms2);
my (@adx,@ady,@tdx,@tdy);
my (@adx_abs,@ady_abs,@adx_rms,@ady_rms);

# unset values
undef(@drao);
undef(@ddeco);
undef(@diffo);
undef(@match);
undef(@redo_f);
undef(@redo_i);
undef(@num_ch);
undef(@fi);
undef(@kfi);

# unset and initialize values
for ($f=2; $f<+$maxf;$f++) {
    $dn[$f]=0;
    $dx[$f]=0.;
    $dy[$f]=0.;
    $dx_abs[$f]=0.;
    $dy_abs[$f]=0.;
    $dx_rms2[$f]=0.;
    $dy_rms2[$f]=0.;
}

# comment this out if using alternate for loop fn
$i=0;

# Correlate sources to first channel, begin offsets calculation
if ($maxf>2) {
$f=1;
# alternate for loop fn - saves memory, runs slower
#for ($fn=$f+1; $fn<+$maxf;$fn++) {
#$i=0;
$j=0;

# loop on boxes defined earlier
for ($nd=$nd_min[$f]; $nd<+$nd_max[$f]+1;$nd++) {
for ($k_ra=$min_rab[$f][$nd]; $k_ra<+$max_rab[$f][$nd]+1;$k_ra++) {
for ($k_dec=$min_decb[$f]; $k_dec<+$max_decb[$f]+1;$k_dec++) {
for ($ns=$ns_min[$f]; $ns<+$ns_max[$f]+1;$ns++) {
 for ($k_w=1; $k_w<+$num_kb[$f][$k_ra][$k_dec][$ns];$k_w++) {

  $k=$kb[$f][$k_ra][$k_dec][$ns][$k_w];
  $i=$i+1;
  $fi[$i]=$f;
  $kfi[$f][$i]=$k;
  # check 4x3 grid of surrounding boxes for matches
  # it is important to check +2/-1 boxes in RA due to bin shear
  for ($kn_ra=$k_ra-1; $kn_ra<+$k_ra+3;$kn_ra++) {
  for ($kn_dec=$k_dec-1; $kn_dec<+$k_dec+2;$kn_dec++) {
   if ($kn_dec<0) { 
    $nsf=1; $ns=($ns+1) % 2; $kn_dec=0; }
   if (($kn_ra<0) or ($kn_ra>$maxb-$kn_dec)) {
    if ($kn_ra==-1) { 
     $kn_ra=$maxb-$kn_dec; }
    if ($kn_ra==$maxb-$kn_dec+1) { 
     $kn_ra=0; }
    if ($kn_ra==$maxb-$kn_dec+2) { next; }
   }
   if (($k_ra<int(($maxb-$k_dec+1)/2)) and ($kn_ra>=int(($maxb-$kn_dec+1)/2))) {
    $ndf=1; $nd=($nd+1) % 2; }
   if (($k_ra>=int(($maxb-$k_dec+1)/2)) and ($kn_ra<int(($maxb-$kn_dec+1)/2))) {
    $ndf=1; $nd=($nd+1) % 2; }
   for ($fn=$f+1; $fn<+$maxf;$fn++) {
    if (($kn_ra>=$min_rab[$fn][$nd]) and ($kn_ra<=$max_rab[$fn][$nd]) and ($kn_dec>=$min_decb[$fn]) and ($kn_dec<=$max_decb[$fn])) {
     for ($kn_w=1; $kn_w<+$num_kb[$fn][$kn_ra][$kn_dec][$ns];$kn_w++) {
      $kn=$kb[$fn][$kn_ra][$kn_dec][$ns][$kn_w];
      $ddec=($dec[$f][$k]-$dec[$fn][$kn])*3600.;
      $dra=$ra[$f][$k]-$ra[$fn][$kn];
      if (abs($dra)>=180.) {
       if ($dra>180.) {
        $dra=$dra-360.; }
       if ($dra<=-180.) {
        $dra=$dra+360.; }
      }
      $dra=$dra*3600.*cos(3.14159*$dec[$f][$k]/180.);
      $diff=($dra**2+$ddec**2)**0.5;
      if ($diff <= $rad) {
       if (not $match[$fn][$kn]) {
        if (not $kfi[$fn][$i]) {
         # no previous match
         $dx[$fn]=$dx[$fn]+$dra;
         $dy[$fn]=$dy[$fn]+$ddec;
         $dx_abs[$fn]=$dx_abs[$fn]+abs($dra);
         $dy_abs[$fn]=$dy_abs[$fn]+abs($ddec);
         $dx_rms2[$fn]=$dx_rms2[$fn]+($dra**2);
         $dy_rms2[$fn]=$dy_rms2[$fn]+($ddec**2);
         $dn[$fn]=$dn[$fn]+1;
         $kfi[$fn][$i]=$kn;
         $match[$fn][$kn]=$i;
         $drao[$fn][$i]=$dra;
         $ddeco[$fn][$i]=$ddec;
         $diffo[$fn][$i]=$diff;
        }
        else {
         if ($diff<$diffo[$fn][$i]) {
          # previous match i<->kn
          $dx[$fn]=$dx[$fn]+$dra-$drao[$fn][$i];
          $dy[$fn]=$dy[$fn]+$ddec-$ddeco[$fn][$i];
          $dx_abs[$fn]=$dx_abs[$fn]+abs($dra)-abs($drao[$fn][$i]);
          $dy_abs[$fn]=$dy_abs[$fn]+abs($ddec)-abs($ddeco[$fn][$i]);
          $dx_rms2[$fn]=$dx_rms2[$fn]+($dra**2)-($drao[$fn][$i]**2);
          $dy_rms2[$fn]=$dy_rms2[$fn]+($ddec**2)-($ddeco[$fn][$i]**2);
          $dn[$fn]=$dn[$fn];
          $match[$fn][$kfi[$fn][$i]]="";
          $kfi[$fn][$i]=$kn;
          $match[$fn][$kn]=$i;
          $drao[$fn][$i]=$dra;
          $ddeco[$fn][$i]=$ddec;
          $diffo[$fn][$i]=$diff;
         }
        }
       }
       else {
        if ($diffo[$fn][$i]) {
         if (($diff<$diffo[$fn][$match[$fn][$kn]]) and ($diff<$diffo[$fn][$i])) {
          # previous match i<->kn and kn to different i
          $dx[$fn]=$dx[$fn]+$dra-$drao[$fn][$match[$fn][$kn]];
          $dy[$fn]=$dy[$fn]+$ddec-$ddeco[$fn][$match[$fn][$kn]];
          $dx_abs[$fn]=$dx_abs[$fn]+abs($dra)-abs($drao[$fn][$match[$fn][$kn]]);
          $dy_abs[$fn]=$dy_abs[$fn]+abs($ddec)-abs($ddeco[$fn][$match[$fn][$kn]]);
          $dx_rms2[$fn]=$dx_rms2[$fn]+($dra**2)-($drao[$fn][$match[$fn][$kn]]**2);
          $dy_rms2[$fn]=$dy_rms2[$fn]+($ddec**2)-($ddeco[$fn][$match[$fn][$kn]]**2);
          $dn[$fn]=$dn[$fn];
          # tag orphaned sources - attempt to re-match later
          $j=$j+1;
          $redo_f[$f][$j]=$fn;
          $redo_i[$f][$j]=$match[$fn][$kn];
          #
          $kfi[$fn][$match[$fn][$kn]]="";
          $match[$fn][$kfi[$fn][$i]]="";
          $drao[$fn][$match[$fn][$kn]]="";
          $ddeco[$fn][$match[$fn][$kn]]="";
          $diffo[$fn][$match[$fn][$kn]]="";
          $kfi[$fn][$i]=$kn;
          $match[$fn][$kn]=$i;
          $drao[$fn][$i]=$dra;
          $ddeco[$fn][$i]=$ddec;
          $diffo[$fn][$i]=$diff;
         }
        }
        else {
         if ($diff<$diffo[$fn][$match[$fn][$kn]]) {
          # previous match kn to different i
          $dx[$fn]=$dx[$fn]+$dra-$drao[$fn][$match[$fn][$kn]];
          $dy[$fn]=$dy[$fn]+$ddec-$ddeco[$fn][$match[$fn][$kn]];
          $dx_abs[$fn]=$dx_abs[$fn]+abs($dra)-abs($drao[$fn][$match[$fn][$kn]]);
          $dy_abs[$fn]=$dy_abs[$fn]+abs($ddec)-abs($ddeco[$fn][$match[$fn][$kn]]);
          $dx_rms2[$fn]=$dx_rms2[$fn]+($dra**2)-($drao[$fn][$match[$fn][$kn]]**2);
          $dy_rms2[$fn]=$dy_rms2[$fn]+($ddec**2)-($ddeco[$fn][$match[$fn][$kn]]**2);
          $dn[$fn]=$dn[$fn];
          # tag orphaned sources - attempt to re-match later
          $j=$j+1;
          $redo_f[$f][$j]=$fn;
          $redo_i[$f][$j]=$match[$fn][$kn];
          #
          $kfi[$fn][$match[$fn][$kn]]="";
          $drao[$fn][$match[$fn][$kn]]="";
          $ddeco[$fn][$match[$fn][$kn]]="";
          $diffo[$fn][$match[$fn][$kn]]="";
          $kfi[$fn][$i]=$kn;
          $match[$fn][$kn]=$i;
          $drao[$fn][$i]=$dra;
          $ddeco[$fn][$i]=$ddec;
          $diffo[$fn][$i]=$diff;
         }
        }
       }
      }
     } #end for loop kn_w
    }
   } #end for loop fn
   if ($ndf) {
    $nd=($nd+1) % 2; undef($ndf); }
   if ($nsf) {
    $ns=($ns+1) % 2; undef($nsf); }
  } #end for loop kn_dec
  } #end for loop kn_ra

 } #end for loop k_w
} #end for loop ns
} #end for loop k_dec
} #end for loop k_ra
} #end for loop nd
$jtot=$j+1;

# Re-match orphaned sources to remaining available sources
# check only those channels from which they were orphaned
# possible, though improbable
for ($j=1; $j<+$jtot;$j++) {
 $fno=$redo_f[$f][$j];
 $io=$redo_i[$f][$j];
 $k=$kfi[$f][$io];
 # set box for orphaned source
 if ($dec[$f][$k]<0) {
  $ns=0;
  $k_dec=int(abs($dec[$f][$k])*3600./$box);
 }
 else {
  $ns=1;
  $k_dec=int($dec[$f][$k]*3600./$box);
 }
 if ($ra[$f][$k]<180) {
  $nd=0; }
 else {
  $nd=1; }
 $k_ra=int($ra[$f][$k]*($maxb-$k_dec+1)/360.);
 # check 4x3 grid of surrounding boxes for matches
 # it is important to check +2/-1 boxes in RA due to bin shear
 for ($kn_ra=$k_ra-1; $kn_ra<+$k_ra+3;$kn_ra++) {
 for ($kn_dec=$k_dec-1; $kn_dec<+$k_dec+2;$kn_dec++) {
  if ($kn_dec<0) { 
   $nsf=1; $ns=($ns+1) % 2; $kn_dec=0; }
  if (($kn_ra<0) or ($kn_ra>$maxb-$kn_dec)) {
   if ($kn_ra==-1) { 
    $kn_ra=$maxb-$kn_dec; }
   if ($kn_ra==$maxb-$kn_dec+1) { 
    $kn_ra=0; }
   if ($kn_ra==$maxb-$kn_dec+2) { next; }
  }
  if (($k_ra<int(($maxb-$k_dec+1)/2)) and ($kn_ra>=int(($maxb-$kn_dec+1)/2))) {
   $ndf=1; $nd=($nd+1) % 2; }
  if (($k_ra>=int(($maxb-$k_dec+1)/2)) and ($kn_ra<int(($maxb-$kn_dec+1)/2))) {
   $ndf=1; $nd=($nd+1) % 2; }
  if (($kn_ra>=$min_rab[$fno][$nd]) and ($kn_ra<=$max_rab[$fno][$nd]) and ($kn_dec>=$min_decb[$fno]) and ($kn_dec<=$max_decb[$fno])) {
   for ($kn_w=1; $kn_w<+$num_kb[$fno][$kn_ra][$kn_dec][$ns];$kn_w++) {
    $kn=$kb[$fno][$kn_ra][$kn_dec][$ns][$kn_w];
    $ddec=($dec[$f][$k]-$dec[$fno][$kn])*3600.;
    $dra=$ra[$f][$k]-$ra[$fno][$kn];
    if (abs($dra)>=180.) {
     if ($dra>180.) {
      $dra=$dra-360.; }
     if ($dra<=-180.) {
      $dra=$dra+360.; }
    }
    $dra=$dra*3600.*cos(3.14159*$dec[$f][$k]/180.);
    $diff=($dra**2+$ddec**2)**0.5;
    if (($diff <= $rad) and (not $match[$fno][$kn])) {
     if (not $kfi[$fno][$io]) {
      # no previous match
      $dx[$fno]=$dx[$fno]+$dra;
      $dy[$fno]=$dy[$fno]+$ddec;
      $dx_abs[$fno]=$dx_abs[$fno]+abs($dra);
      $dy_abs[$fno]=$dy_abs[$fno]+abs($ddec);
      $dx_rms2[$fno]=$dx_rms2[$fno]+($dra**2);
      $dy_rms2[$fno]=$dy_rms2[$fno]+($ddec**2);
      $dn[$fno]=$dn[$fno]+1;
      $kfi[$fno][$io]=$kn;
      $match[$fno][$kn]=$io;
      $drao[$fno][$io]=$dra;
      $ddeco[$fno][$io]=$ddec;
      $diffo[$fno][$io]=$diff;
     }
     else {
      if ($diff<$diffo[$fno][$io]) {
       # previous match i<->kn
       $dx[$fno]=$dx[$fno]+$dra-$drao[$fno][$io];
       $dy[$fno]=$dy[$fno]+$ddec-$ddeco[$fno][$io];
       $dx_abs[$fno]=$dx_abs[$fno]+abs($dra)-abs($drao[$fno][$io]);
       $dy_abs[$fno]=$dy_abs[$fno]+abs($ddec)-abs($ddeco[$fno][$io]);
       $dx_rms2[$fno]=$dx_rms2[$fno]+($dra**2)-($drao[$fno][$io]**2);
       $dy_rms2[$fno]=$dy_rms2[$fno]+($ddec**2)-($ddeco[$fno][$io]**2);
       $dn[$fno]=$dn[$fno];
       $match[$fno][$kfi[$fno][$io]]="";
       $kfi[$fno][$io]=$kn;
       $match[$fno][$kn]=$io;
       $drao[$fno][$io]=$dra;
       $ddeco[$fno][$io]=$ddec;
       $diffo[$fno][$io]=$diff;
      }
     }
    }
   } #end for loop kn_w
  }
  if ($ndf) {
   $nd=($nd+1) % 2; undef($ndf); }
  if ($nsf) {
   $ns=($ns+1) % 2; undef($nsf); }
 } #end for loop kn_dec
 } #end for loop kn_ra
} #end for loop j

# unset values
undef(@drao);
undef(@ddeco);
undef(@diffo);
undef(@redo_f);
undef(@redo_i);

#} #end alternate for loop fn - saves memory, runs slower
$num_ch[$f]=$i;
} #end if
#end offsets calculation

# Correlate remaining sources within remaining channels
# for photometry file only - not used for offset calculation
if ($maxf>3) {
for ($f=2; $f<+$maxf-1;$f++) {
# alternate for loop fn - saves memory, runs slower
#for ($fn=$f+1; $fn<+$maxf;$fn++) {
$i=$num_ch[$f-1];
$j=0;

# loop on boxes defined earlier
for ($nd=$nd_min[$f]; $nd<+$nd_max[$f]+1;$nd++) {
for ($k_ra=$min_rab[$f][$nd]; $k_ra<+$max_rab[$f][$nd]+1;$k_ra++) {
for ($k_dec=$min_decb[$f]; $k_dec<+$max_decb[$f]+1;$k_dec++) {
for ($ns=$ns_min[$f]; $ns<+$ns_max[$f]+1;$ns++) {
 for ($k_w=1; $k_w<+$num_kb[$f][$k_ra][$k_dec][$ns];$k_w++) {

  $k=$kb[$f][$k_ra][$k_dec][$ns][$k_w];
  if (not $match[$f][$k]) {
   $i=$i+1;
   $fi[$i]=$f;
   $kfi[$f][$i]=$k;
   # check 4x3 grid of surrounding boxes for matches
   # it is important to check +2/-1 boxes in RA due to bin shear
   for ($kn_ra=$k_ra-1; $kn_ra<+$k_ra+3;$kn_ra++) {
   for ($kn_dec=$k_dec-1; $kn_dec<+$k_dec+2;$kn_dec++) {
    if ($kn_dec<0) { 
     $nsf=1; $ns=($ns+1) % 2; $kn_dec=0; }
    if (($kn_ra<0) or ($kn_ra>$maxb-$kn_dec)) {
     if ($kn_ra==-1) { 
      $kn_ra=$maxb-$kn_dec; }
     if ($kn_ra==$maxb-$kn_dec+1) { 
      $kn_ra=0; }
     if ($kn_ra==$maxb-$kn_dec+2) { next; }
    }
    if (($k_ra<int(($maxb-$k_dec+1)/2)) and ($kn_ra>=int(($maxb-$kn_dec+1)/2))) {
     $ndf=1; $nd=($nd+1) % 2; }
    if (($k_ra>=int(($maxb-$k_dec+1)/2)) and ($kn_ra<int(($maxb-$kn_dec+1)/2))) {
     $ndf=1; $nd=($nd+1) % 2; }
    for ($fn=$f+1; $fn<+$maxf;$fn++) {
     if (($kn_ra>=$min_rab[$fn][$nd]) and ($kn_ra<=$max_rab[$fn][$nd]) and ($kn_dec>=$min_decb[$fn]) and ($kn_dec<=$max_decb[$fn])) {
      for ($kn_w=1; $kn_w<+$num_kb[$fn][$kn_ra][$kn_dec][$ns];$kn_w++) {
       $kn=$kb[$fn][$kn_ra][$kn_dec][$ns][$kn_w];
       $ddec=($dec[$f][$k]-$dec[$fn][$kn])*3600.;
       $dra=$ra[$f][$k]-$ra[$fn][$kn];
       if (abs($dra)>=180.) {
        if ($dra>180.) {
         $dra=$dra-360.; }
        if ($dra<=-180.) {
         $dra=$dra+360.; }
       }
       $dra=$dra*3600.*cos(3.14159*$dec[$f][$k]/180.);
       $diff=($dra**2+$ddec**2)**0.5;
       if ($diff <= $rad) {
        if (not $match[$fn][$kn]) {
         if (not $kfi[$fn][$i]) {
          # no previous match
          $kfi[$fn][$i]=$kn;
          $match[$fn][$kn]=$i;
          $drao[$fn][$i]=$dra;
          $ddeco[$fn][$i]=$ddec;
          $diffo[$fn][$i]=$diff;
         }
         else {
          if ($diff<$diffo[$fn][$i]) {
           # previous match i<->kn
           $match[$fn][$kfi[$fn][$i]]="";
           $kfi[$fn][$i]=$kn;
           $match[$fn][$kn]=$i;
           $drao[$fn][$i]=$dra;
           $ddeco[$fn][$i]=$ddec;
           $diffo[$fn][$i]=$diff;
          }
         }
        }
        else {
         if ($fi[$match[$fn][$kn]] == $f) {
          if ($diffo[$fn][$i]) {
           if (($diff<$diffo[$fn][$match[$fn][$kn]]) and ($diff<$diffo[$fn][$i])) {
            # previous match i<->kn and kn to different i
            # tag orphaned sources - attempt to re-match later
            $j=$j+1;
            $redo_f[$f][$j]=$fn;
            $redo_i[$f][$j]=$match[$fn][$kn];
            #
            $kfi[$fn][$match[$fn][$kn]]="";
            $match[$fn][$kfi[$fn][$i]]="";
            $drao[$fn][$match[$fn][$kn]]="";
            $ddeco[$fn][$match[$fn][$kn]]="";
            $diffo[$fn][$match[$fn][$kn]]="";
            $kfi[$fn][$i]=$kn;
            $match[$fn][$kn]=$i;
            $drao[$fn][$i]=$dra;
            $ddeco[$fn][$i]=$ddec;
            $diffo[$fn][$i]=$diff;
           }
          }
          else {
           if ($diff<$diffo[$fn][$match[$fn][$kn]]) {
            # previous match kn to different i
            # tag orphaned sources - attempt to re-match later
            $j=$j+1;
            $redo_f[$f][$j]=$fn;
            $redo_i[$f][$j]=$match[$fn][$kn];
            #
            $kfi[$fn][$match[$fn][$kn]]="";
            $drao[$fn][$match[$fn][$kn]]="";
            $ddeco[$fn][$match[$fn][$kn]]="";
            $diffo[$fn][$match[$fn][$kn]]="";
            $kfi[$fn][$i]=$kn;
            $match[$fn][$kn]=$i;
            $drao[$fn][$i]=$dra;
            $ddeco[$fn][$i]=$ddec;
            $diffo[$fn][$i]=$diff;
           }
          }
         }
        }
       }
      } #end for loop kn_w
     }
    } #end for loop fn
    if ($ndf) {
     $nd=($nd+1) % 2; undef($ndf); }
    if ($nsf) {
     $ns=($ns+1) % 2; undef($nsf); }
   } #end for loop kn_dec
   } #end for loop kn_ra
  }

 } #end for loop k_w
} #end for loop ns
} #end for loop k_dec
} #end for loop k_ra
} #end for loop nd
$jtot=$j+1;

# Re-match orphaned sources to remaining available sources
# check only those channels from which they were orphaned
# possible, though improbable
for ($j=1; $j<+$jtot;$j++) {
 $fno=$redo_f[$f][$j];
 $io=$redo_i[$f][$j];
 $k=$kfi[$f][$io];
 # set box for orphaned source
 if ($dec[$f][$k]<0) {
  $ns=0;
  $k_dec=int(abs($dec[$f][$k])*3600./$box);
 }
 else {
  $ns=1;
  $k_dec=int($dec[$f][$k]*3600./$box);
 }
 if ($ra[$f][$k]<180) {
  $nd=0; }
 else {
  $nd=1; }
 $k_ra=int($ra[$f][$k]*($maxb-$k_dec+1)/360.);
 # check 4x3 grid of surrounding boxes for matches
 # it is important to check +2/-1 boxes in RA due to bin shear
 for ($kn_ra=$k_ra-1; $kn_ra<+$k_ra+3;$kn_ra++) {
 for ($kn_dec=$k_dec-1; $kn_dec<+$k_dec+2;$kn_dec++) {
  if ($kn_dec<0) { 
   $nsf=1; $ns=($ns+1) % 2; $kn_dec=0; }
  if (($kn_ra<0) or ($kn_ra>$maxb-$kn_dec)) {
   if ($kn_ra==-1) { 
    $kn_ra=$maxb-$kn_dec; }
   if ($kn_ra==$maxb-$kn_dec+1) { 
    $kn_ra=0; }
   if ($kn_ra==$maxb-$kn_dec+2) { next; }
  }
  if (($k_ra<int(($maxb-$k_dec+1)/2)) and ($kn_ra>=int(($maxb-$kn_dec+1)/2))) {
   $ndf=1; $nd=($nd+1) % 2; }
  if (($k_ra>=int(($maxb-$k_dec+1)/2)) and ($kn_ra<int(($maxb-$kn_dec+1)/2))) {
   $ndf=1; $nd=($nd+1) % 2; }
  if (($kn_ra>=$min_rab[$fno][$nd]) and ($kn_ra<=$max_rab[$fno][$nd]) and ($kn_dec>=$min_decb[$fno]) and ($kn_dec<=$max_decb[$fno])) {
   for ($kn_w=1; $kn_w<+$num_kb[$fno][$kn_ra][$kn_dec][$ns];$kn_w++) {
    $kn=$kb[$fno][$kn_ra][$kn_dec][$ns][$kn_w];
    $ddec=($dec[$f][$k]-$dec[$fno][$kn])*3600.;
    $dra=$ra[$f][$k]-$ra[$fno][$kn];
    if (abs($dra)>=180.) {
     if ($dra>180.) {
      $dra=$dra-360.; }
     if ($dra<=-180.) {
      $dra=$dra+360.; }
    }
    $dra=$dra*3600.*cos(3.14159*$dec[$f][$k]/180.);
    $diff=($dra**2+$ddec**2)**0.5;
    if (($diff <= $rad) and (not $match[$fno][$kn])) {
     if (not $kfi[$fno][$io]) {
      # no previous match
      $kfi[$fno][$io]=$kn;
      $match[$fno][$kn]=$io;
      $drao[$fno][$io]=$dra;
      $ddeco[$fno][$io]=$ddec;
      $diffo[$fno][$io]=$diff;
     }
     else {
      if ($diff<$diffo[$fno][$io]) {
       # previous match i<->kn
       $match[$fno][$kfi[$fno][$io]]="";
       $kfi[$fno][$io]=$kn;
       $match[$fno][$kn]=$io;
       $drao[$fno][$io]=$dra;
       $ddeco[$fno][$io]=$ddec;
       $diffo[$fno][$io]=$diff;
      }
     }
    }
   } #end for loop kn_w
  }
  if ($ndf) {
   $nd=($nd+1) % 2; undef($ndf); }
  if ($nsf) {
   $ns=($ns+1) % 2; undef($nsf); }
 } #end for loop kn_dec
 } #end for loop kn_ra
} #end for loop j

# unset values
undef(@drao);
undef(@ddeco);
undef(@diffo);
undef(@redo_f);
undef(@redo_i);

#} #end alternate for loop fn - saves memory, runs slower
$num_ch[$f]=$i;

} #end for loop f
} #end if

# Add remaining non-matched sources to list from last channel
$f=$maxf-1;

# loop on boxes defined earlier
for ($nd=$nd_min[$f]; $nd<+$nd_max[$f]+1;$nd++) {
for ($k_ra=$min_rab[$f][$nd]; $k_ra<+$max_rab[$f][$nd]+1;$k_ra++) {
for ($k_dec=$min_decb[$f]; $k_dec<+$max_decb[$f]+1;$k_dec++) {
for ($ns=$ns_min[$f]; $ns<+$ns_max[$f]+1;$ns++) {
 for ($k_w=1; $k_w<+$num_kb[$f][$k_ra][$k_dec][$ns];$k_w++) {
  $k=$kb[$f][$k_ra][$k_dec][$ns][$k_w];
  if (not $match[$f][$k]) {
   $i=$i+1;
   $fi[$i]=$f;
   $kfi[$f][$i]=$k;
  }
 } #end for loop k_w
} #end for loop ns
} #end for loop k_dec
} #end for loop k_ra
} #end for loop nd

# set total number of "unique" sources
$itot=$i+1;


## ######################################
## Print summary to terminal and log file
## ######################################

# Must always initialize adx and ady for chs. 2, 3, and 4
for ($c=2; $c<+5;$c++) {
    $adx[$c]=0.;
    $ady[$c]=0.;
}

my $file3=join '',($output,".log");

open (FILE3,"> $file3") or die "cannot open file to write to";

# Print settings
printf FILE3 "USAGE:  perl $pname [--help][--ver]\n";
printf FILE3 "VERSION:  $VERSION\n";
printf FILE3 "\n";
printf FILE3 "Tolerance for average dRA and dDec = $toler\"\n";
printf FILE3 "Maximum number of iterations = $iter\n";
printf FILE3 "Matching radius = $rad\"\n";
if ($bsam) {
 printf FILE3 "Search box size scale = $bsam\'\n"; }
else {
 printf FILE3 "Search box size scale = $boxscale * matching radius\n"; }
printf FILE3 "Aggressive Mode?: $a_label\n";
printf FILE3 "Maximum photometric error = $max_err mag.\n";
printf FILE3 "Minimum signal-to-noise = $min_s2n\n";
printf FILE3 "Red source color cut: ch1-ch2 >= $color\n";
printf FILE3 "Photometry file root: $phot\n";
printf FILE3 "WCS coordinate file root: $wcs\n";
printf FILE3 "Correlated photometry file root: $output\n";
printf FILE3 "\n";

printf FILE3 "Register WCS to channel 1 (shifts are in \")\n";
printf FILE3 "Exit after step $h:\n";
printf FILE3 "\n";
printf FILE3 "%2s%1s%9s%10s%1s%7s%7s%1s%7s%7s%1s%8s%9s%1s%8s\n","ch","","RA shift","Dec shift","","dRA","dDec","","|dRA|","|dDec|","","rms dRA","rms dDec","","matches";
print "\n";
print "Register WCS to channel 1 (shifts are in \")\n";
print "step $h:\n";
print "\n";
print "ch  RA shift Dec shift     dRA   dDec   |dRA| |dDec|  rms dRA rms dDec  matches\n";

for ($f=2; $f<+$maxf;$f++) {
    if ($dn[$f] > 0) {
     $adx[$f]=$dx[$f]/$dn[$f];
     $ady[$f]=$dy[$f]/$dn[$f];
     $adx_abs[$f]=$dx_abs[$f]/$dn[$f];
     $ady_abs[$f]=$dy_abs[$f]/$dn[$f];;
     $adx_rms[$f]=($dx_rms2[$f]/$dn[$f])**0.5;
     $ady_rms[$f]=($dy_rms2[$f]/$dn[$f])**0.5;
     $tdx[$f]=$adx[$f]+$dx0[$f];
     $tdy[$f]=$ady[$f]+$dy0[$f];
     $dx0[$f]=$tdx[$f];
     $dy0[$f]=$tdy[$f];
     $adx[$f]=sprintf "%7.3f",$adx[$f];
     $ady[$f]=sprintf "%6.3f",$ady[$f];
     $adx_abs[$f]=sprintf "%7.3f",$adx_abs[$f];
     $ady_abs[$f]=sprintf "%6.3f",$ady_abs[$f];
     $adx_rms[$f]=sprintf "%8.3f",$adx_rms[$f];
     $ady_rms[$f]=sprintf "%8.3f",$ady_rms[$f];
     $tdx[$f]=sprintf "%9.3f",$tdx[$f];
     $tdy[$f]=sprintf "%9.3f",$tdy[$f];
     $dn[$f]=sprintf "%8i",$dn[$f];
     printf FILE3 "%2i%1s%9.3f%1s%9.3f%1s%7.3f%1s%6.3f%1s%7.3f%1s%6.3f%1s%8.3f%1s%8.3f%1s%8i\n",$cf[$f],"",$tdx[$f],"",$tdy[$f],"",$adx[$f],"",$ady[$f],"",$adx_abs[$f],"",$ady_abs[$f],"",$adx_rms[$f],"",$ady_rms[$f],"",$dn[$f];
     print " $cf[$f] $tdx[$f] $tdy[$f] $adx[$f] $ady[$f] $adx_abs[$f] $ady_abs[$f] $adx_rms[$f] $ady_rms[$f] $dn[$f]\n";
    }
    else {
     print "0 matching sources, Channel ",$f," not registered properly\n";
     printf FILE3 "0 matching sources, Channel $f not registered properly\n";
     $adx[$f]=0.;
     $ady[$f]=0.;
     $adx_abs[$f]=0.;
     $ady_abs[$f]=0.;
     $adx_rms[$f]=0.;
     $ady_rms[$f]=0.;
     $tdx[$f]=$adx[$f]+$dx0[$f];
     $tdy[$f]=$ady[$f]+$dy0[$f];
     $dx0[$f]=$tdx[$f];
     $dy0[$f]=$tdy[$f];
     $adx[$f]=sprintf "%7.3f",$adx[$f];
     $ady[$f]=sprintf "%6.3f",$ady[$f];
     $adx_abs[$f]=sprintf "%7.3f",$adx_abs[$f];
     $ady_abs[$f]=sprintf "%6.3f",$ady_abs[$f];
     $adx_rms[$f]=sprintf "%8.3f",$adx_rms[$f];
     $ady_rms[$f]=sprintf "%8.3f",$ady_rms[$f];
     $tdx[$f]=sprintf "%9.3f",$tdx[$f];
     $tdy[$f]=sprintf "%9.3f",$tdy[$f];
     $dn[$f]=sprintf "%8i",$dn[$f];
     printf FILE3 "%2i%1s%9.3f%1s%9.3f%1s%7.3f%1s%6.3f%1s%7.3f%1s%6.3f%1s%8.3f%1s%8.3f%1s%8i\n",$cf[$f],"",$tdx[$f],"",$tdy[$f],"",$adx[$f],"",$ady[$f],"",$adx_abs[$f],"",$ady_abs[$f],"",$adx_rms[$f],"",$ady_rms[$f],"",$dn[$f];
     print " $cf[$f] $tdx[$f] $tdy[$f] $adx[$f] $ady[$f] $adx_abs[$f] $ady_abs[$f] $adx_rms[$f] $ady_rms[$f] $dn[$f]\n";
    }
} #end for loop
close FILE3;

$h=$h+1;
$maxd=max_abs($adx[2],$adx[3],$adx[4],$ady[2],$ady[3],$ady[4]);

undef(@match);

} #end while loop
print "\n";

undef(@kb);
undef(@num_kb);


## ############################################################################
## End iterations
## ############################################################################

## ############################################################################
## Convert WCS coordinates back to sexagesimal
## ############################################################################

my @kc;

for ($i=1; $i<+$itot;$i++) {

    # copy tags
    $kc[$i]=$kfi[$fi[$i]][$i];

    my ($moo1,$moo2);

    # convert RA to hh:mm:ss
    $moo1=$ra[$fi[$i]][$kc[$i]]/15.;
    ($hr[$fi[$i]][$kc[$i]],$moo2)=split(/\./,$moo1);
    $moo1=60.*($moo1-$hr[$fi[$i]][$kc[$i]]);
    ($min[$fi[$i]][$kc[$i]],$moo2)=split(/\./,$moo1);
    $sec[$fi[$i]][$kc[$i]]=60.*($moo1-$min[$fi[$i]][$kc[$i]]);
    if ($sec[$fi[$i]][$kc[$i]] < 0.) {
     $sec[$fi[$i]][$kc[$i]]=0.; }
    if ($min[$fi[$i]][$kc[$i]] < 10) {
     $min[$fi[$i]][$kc[$i]]=join "",("0",$min[$fi[$i]][$kc[$i]]); }
    $sec[$fi[$i]][$kc[$i]]=sprintf "%6.3f",$sec[$fi[$i]][$kc[$i]];
    if ($sec[$fi[$i]][$kc[$i]] < 10) {
     $sec[$fi[$i]][$kc[$i]]=sprintf "%5.3f",$sec[$fi[$i]][$kc[$i]];
     $sec[$fi[$i]][$kc[$i]]=join "",("0",$sec[$fi[$i]][$kc[$i]]);
    }
    else {
     $sec[$fi[$i]][$kc[$i]]=sprintf "%6.3f",$sec[$fi[$i]][$kc[$i]]; }

    # convert Dec to dd:mm:ss
    ($deg[$fi[$i]][$kc[$i]],$moo2)=split(/\./,$dec[$fi[$i]][$kc[$i]]);
    if ($dec[$fi[$i]][$kc[$i]]<0) {
     $moo1=-60.*($dec[$fi[$i]][$kc[$i]]-$deg[$fi[$i]][$kc[$i]]); }
    else {
     $moo1=60.*($dec[$fi[$i]][$kc[$i]]-$deg[$fi[$i]][$kc[$i]]); }
    ($am[$fi[$i]][$kc[$i]],$moo2)=split(/\./,$moo1);
    $as[$fi[$i]][$kc[$i]]=60.*($moo1-$am[$fi[$i]][$kc[$i]]);
    if ($as[$fi[$i]][$kc[$i]] < 0.) {
     $as[$fi[$i]][$kc[$i]]=0.; }
    if ($am[$fi[$i]][$kc[$i]] < 10) {
     $am[$fi[$i]][$kc[$i]]=join "",("0",$am[$fi[$i]][$kc[$i]]); }
    $as[$fi[$i]][$kc[$i]]=sprintf "%6.3f",$as[$fi[$i]][$kc[$i]];
    if ($as[$fi[$i]][$kc[$i]] < 10) {
     $as[$fi[$i]][$kc[$i]]=sprintf "%4.2f",$as[$fi[$i]][$kc[$i]];
     $as[$fi[$i]][$kc[$i]]=join "",("0",$as[$fi[$i]][$kc[$i]]);
    }
    else {
     $as[$fi[$i]][$kc[$i]]=sprintf "%5.2f",$as[$fi[$i]][$kc[$i]]; }

} #end for loop i


## ############################################################################
## Output correlated photometry files
## ############################################################################

my $file4=join '',($output,".all");
my $file4r=join '',($output,".all.reg");
my $file5=join '',($output,".red");
my $file5r=join '',($output,".red.reg");

open (FILE4,"> $file4") or die "$pname: cannot open file to write to\n";
open (FILE4r,"> $file4r") or die "$pname: cannot open file to write to\n";
printf FILE4 "%1s%9s%3s%11s%3s%9s%7s%6s%7s%6s%7s%6s%7s\n","#","","RA","","Dec","","ch $cf[1]","","ch $cf[2]","","ch $cf[3]","","ch $cf[4]";
printf FILE4 "%1s%5s%4s%3s%3s%4s%4s%3s%3s%3s%7s%6s%7s%6s%7s%6s%7s%6s\n","#","","hh","mm","ss","","dd","mm","ss","","mag","sig","mag","sig","mag","sig","mag","sig";
printf FILE4 "\n";

if (($fc[1]) and ($fc[2])) {
    open (FILE5,"> $file5") or die "$pname: cannot open file to write to\n";
    open (FILE5r,"> $file5r") or die "$pname: cannot open file to write to\n";
    printf FILE5 "%1s%9s%3s%11s%3s%9s%7s%6s%7s%6s%7s%6s%7s\n","#","","RA","","Dec","","ch $cf[1]","","ch $cf[2]","","ch $cf[3]","","ch $cf[4]";
    printf FILE5 "%1s%5s%4s%3s%3s%4s%4s%3s%3s%3s%7s%6s%7s%6s%7s%6s%7s%6s\n","#","","hh","mm","ss","","dd","mm","ss","","mag","sig","mag","sig","mag","sig","mag","sig";
    printf FILE5 "\n";
}

my (@k1,@k2,@k3,@k4);

for ($i=1; $i<+$itot;$i++) {

    # copy tags and set values for blank entries
    if ($kfi[1][$i]) {
     $k1[$i]=$kfi[1][$i]; }
    else {
     $k1[$i]=$itot+1;
     $mag[1][$k1[$i]]="";
     $sig[1][$k1[$i]]="";
    }
    if ($kfi[2][$i]) {
     $k2[$i]=$kfi[2][$i]; }
    else {
     $k2[$i]=$itot+1;
     $mag[2][$k2[$i]]="";
     $sig[2][$k2[$i]]="";
    }
    if ($kfi[3][$i]) {
     $k3[$i]=$kfi[3][$i]; }
    else {
     $k3[$i]=$itot+1;
     $mag[3][$k3[$i]]="";
     $sig[3][$k3[$i]]="";
    }
    if ($kfi[4][$i]) {
     $k4[$i]=$kfi[4][$i]; }
    else {
     $k4[$i]=$itot+1;
     $mag[4][$k4[$i]]="";
     $sig[4][$k4[$i]]="";
    }

    printf FILE4 "%6s%4s%3s%7s%4s%3s%6s%7s%6s%7s%6s%7s%6s%7s%6s\n",$i,$hr[$fi[$i]][$kc[$i]],$min[$fi[$i]][$kc[$i]],$sec[$fi[$i]][$kc[$i]],$deg[$fi[$i]][$kc[$i]],$am[$fi[$i]][$kc[$i]],$as[$fi[$i]][$kc[$i]],$mag[1][$k1[$i]],$sig[1][$k1[$i]],$mag[2][$k2[$i]],$sig[2][$k2[$i]],$mag[3][$k3[$i]],$sig[3][$k3[$i]],$mag[4][$k4[$i]],$sig[4][$k4[$i]];
    printf FILE4r "fk5;circle($ra[$fi[$i]][$kc[$i]],$dec[$fi[$i]][$kc[$i]],$rad\")\n";

    if (($fc[1]) and ($fc[2])) {
     my $c12="na";
     if (($mag[1][$k1[$i]]) and ($mag[2][$k2[$i]])) {
      $c12=$mag[1][$k1[$i]]-$mag[2][$k2[$i]]; }
     if ($c12 ne "na") {
      if ($c12 >= $color) {
       printf FILE5 "%6s%4s%3s%7s%4s%3s%6s%7s%6s%7s%6s%7s%6s%7s%6s\n",$i,$hr[$fi[$i]][$kc[$i]],$min[$fi[$i]][$kc[$i]],$sec[$fi[$i]][$kc[$i]],$deg[$fi[$i]][$kc[$i]],$am[$fi[$i]][$kc[$i]],$as[$fi[$i]][$kc[$i]],$mag[1][$k1[$i]],$sig[1][$k1[$i]],$mag[2][$k2[$i]],$sig[2][$k2[$i]],$mag[3][$k3[$i]],$sig[3][$k3[$i]],$mag[4][$k4[$i]],$sig[4][$k4[$i]];
       printf FILE5r "fk5;circle($ra[$fi[$i]][$kc[$i]],$dec[$fi[$i]][$kc[$i]],$rad\")\n";
      }
     }
    }

} #end for loop
close FILE4;
close FILE4r;

if (($fc[1]) and ($fc[2])) {
    close FILE5;
    close FILE5r;
}


## ############################################################################
## End
## ############################################################################

#system "date";

exit 0;
