## ############################################################################
## fixup - prepares directory structure and files for IRACproc
##
## (c) 2010 IASTATE, SAO, Massimo Marengo - mmarengo@iastate.edu 
##                        and Mike Schuster - mschuster@cfa.harvard.edu
##
## ############################################################################

=head1 NAME

perl fixup.pl - prepares directory structure and files for IRACproc


=head1 SYNOPSIS

 % perl fixup.pl [--help][--ver][--aorlist file]
                 [--link][--bcdpath string][--prefix string]
                 [--pmasks mmmyy][--mypmask_dir string][--subarray]
                  proc_dir [aor1 aor2 ... aorN]


=head1 COMMAND LINE OPTIONS

 --help
   Print usage information and exit.

 --ver
   Print current version and exit.

 --aorlist file
   File with a list of AOR directories to process

 --link
   Set this flag if you want to symbolically link to BCD data in AOR dirs,
   Warning: The paths for each AOR dir must be absolute.
   not set by default

 --bcdpath string
   Set this when processing data whose directory structure has been modified 
   from the standard SSC AOR directory format. This specifies the path within 
   the data directory to where the BCD data are located (e.g. /). Note, that 
   this assumes that any modification put data from all IRAC bands in a 
   single directory.
   not set by default - in this case bcdpath defaults to 'ch<CH>/bcd'

 --prefix string
   Prefix appended to _bcd.fits and _bunc.fits files during preprocessing. 
   This should be everything that comes before 'I<CH>_...' in the filenames.
   default is 'SPITZER_', set to 'null' for ''.

 --pmasks mmmyy
   Set of Pmasks, as distributed by the SSC, specified by month and year
   available options: sep07 (default), jun06, dec05, jun05, feb05, jan05,
   oct04, jun04, mar04

 --my_pmask_dir string
   Alternative pmasks directory. The default name for the pmasks are
   I<CH>_pmask.fits unless specified with the my_pmask option.

 --my_pmask string
   Name of alternative pmask files. The default is I<CH>_pmask.fits.

 --subarray
   Set this flag if you are processing data taken in sub-array mode
   not set by default

 proc_dir
   name of the top level directory for IRACproc processing

 aor1 aor2 ... aorN
   optional list of AOR directories to process (either this list or the --aor 
   file option should be given)


=head1 DESCRIPTION

This script creates the directory structure for IRACproc and copies images 
from all four IRAC channels into the reduce/ directory. This script sets the 
location of IRACproc and other processing parameters in several IRACproc 
configuration files. During the copy stage the script reformats the bcd data 
delivered by the SSC since summer of 2004.

The directory structure created is:

 proc_dir/

  Lists/
  analysis/
  reduce/

The file names are changed in the following ways:

 "prefix" is removed
 EXPID is padded with up to four "0"s
 DCENUM is kept to one digit

The list of AOR directories to process should be a list of related data sets, 
that is, one or more AORs of the same field. This list can be given in one of 
two ways: the aorlist file (aor.lis) listing each AOR directory on a separate 
line (given priority) or on the command line as "aor1 aor2 ... aorN". Both 
relative and absolute paths are acceptable. 


=head1 CHANGES

 4.7 - Added bimsk to list of files to copy (Apr 1, 2010)
 4.6 - Added my_pmask and my_pmask_dir options (Oct 15, 2008)
 4.5 - Updated Pmasks (Sep 2, 2008)
 4.4 - Removed pixel size to mkmopex_bcd (and its sub-routines)
       Updated w/ new Pmasks from SSC (Apr 25, 2007)
 4.3 - added options 'bcdpath' and 'prefix' to integrate SPITZER/IRAC
       data preprocessed with other contributed software (Sept 18, 2006)
 4.2 - removed options: fif, CROTA2, THRESH_OPTION, MIN_PIX_NUM
       now options for mkmopex_bcd
       (Apr 27, 2006)
 4.1 - added option to symbolically link to BCD data in AOR dirs
       (Mar 30, 2006)
 4.0 - added several options that help to customize the configuration
       files for each specific data set
       added functionality of deprecated setup.pl script
       (Feb 16, 2006)
 3.0 - version for IRACproc 3.2, added command line options, increased 
       capacity (no known limit to no. of files), no longer use README 
       supplied by SSC, now runs on only one related group of AORs instead
       of on multiple unrelated AORs (Jun 20, 2005)
 2.1 - increase capacity for number of files that script can handle 
       (at least 6144), keep dcenum to one digit (Feb 8, 2005)
 2.0 - take out 'SPITZER_', pad expid by up to 2 '0's (Sept 22, 2004)
 1.0 - original version (fixdirs) by Brian M. Patten, SAO


=head1 AUTHOR

 (c) 2008 M. Marengo (mmarengo@cfa.harvard.edu)
          and Mike Schuster (mschuster@cfa.harvard.edu)
 (c) 2008 SAO - Smithsonian Astrophysical Observatory


=head1 SEE ALSO

 multi_fixup.pl(1);


=head1 BUGS

Report bugs to Mike Schuster, mschuster@cfa.harvard.edu


=cut

## ############################################################################
## Set Modules and version number
## ############################################################################

use warnings;
use strict;
use diagnostics;
use Getopt::Long;
use Tie::File;

# Set number version
my $VERSION = '4.7';           # Script version number
my $pname   = 'fixup.pl';      # Script name


## ############################################################################
## Read input options
## ############################################################################

my ($help,$ver,$aorlist,$link,$bcdpath,$prefix);
my ($pmasks,$subarray,$my_pmask,$my_pmask_dir);

my %optctl = (help            => \$help,
              ver             => \$ver,
              aorlist         => \$aorlist,
              link            => \$link,
              bcdpath         => \$bcdpath,
              prefix          => \$prefix,
              pmasks          => \$pmasks,
	      my_pmask        => \$my_pmask,
	      my_pmask_dir    => \$my_pmask_dir,
              subarray        => \$subarray,
             );

&GetOptions(\%optctl,
            "help",
            "ver",
            "aorlist=s",
            "link",
            "bcdpath=s",
            "prefix=s",
            "pmasks=s",
	    "my_pmask=s",
	    "my_pmask_dir=s",
            "subarray",
           );


## ############################################################################
## Process input options
## ############################################################################

# Print help and exit
my $usg = "USAGE:  perl $pname [--help][--ver][--aorlist file]\n"
         ."                        [--link][--bcdpath string][--prefix string]\n"
         ."                        [--pmasks mmmyy][--my_pmask string][--my_pmask_dir string][--subarray]\n"
         ."                         proc_dir [aor1 aor2 ... aorN]\n";

if ($help) {
  print STDOUT <<HELP;

  $pname version $VERSION

  Prepare directory structure and files for IRACproc

  $usg

HELP
  exit 0;
}

# Print version and exit
if ($ver) { 
  print "$pname: version $VERSION\n"; 
  exit 0;
}

# Check validity of options
if ($link) {
  print "$pname: WARNING: To link files, the paths for each AOR dir must be absolute\n"; }

if (not $prefix) { $prefix = "SPITZER_"; }
if ($prefix eq 'null') { $prefix = ""; }

# Check the 'aor' option
my @dirlist;
if ($aorlist) {
  open(AOR,$aorlist) or die "$pname: cannot open file $aorlist\n";
  foreach (<AOR>) {
	if (m/^ *$/) { next; }
	if (m/^ *\#/) { next; }
	chomp;
	push @dirlist,$_;
  }
  close(AOR);
}

if (not $pmasks) { $pmasks = "sep07"; }
else {
  if (($pmasks ne 'sep07') and ($pmasks ne 'jun06') and ($pmasks ne 'dec05') and ($pmasks ne 'jun05') and ($pmasks ne 'feb05') and ($pmasks ne 'jan05') and ($pmasks ne 'oct04') and ($pmasks ne 'jun04') and ($pmasks ne 'mar04')) {
    print "$pname: ERROR: pmasks must be one of the following:\n";
    print "          jun06, dec05, jun05, feb05, jan05, oct04, jun04, mar04\n";
    exit 0;
  }
}
my $PMASKFILE = "chan<CH>_".$pmasks."_pmask.fits";
if ($subarray) {
  $PMASKFILE = "chan<CH>_subarray_".$pmasks."_pmask.fits"; }
my $DIR_PMASK = "september07_pmasks";
if ($pmasks eq 'jun06') {
  $DIR_PMASK = "june06_pmasks"; }
if ($pmasks eq 'dec05') {
  $DIR_PMASK = "december05_pmasks"; }
if ($pmasks eq 'jun05') {
  $DIR_PMASK = "june05_pmasks"; }
if ($pmasks eq 'feb05') {
  $DIR_PMASK = "february05_pmasks"; }
if ($pmasks eq 'jan05') {
  $DIR_PMASK = "january05_pmasks"; }
if ($pmasks eq 'oct04') {
  $DIR_PMASK = "october04_pmasks"; }
if ($pmasks eq 'jun04') {
  $DIR_PMASK = "june04_pmasks"; }
if ($pmasks eq 'mar04') {
  $DIR_PMASK = "march04_pmasks"; }
if ($my_pmask_dir) { $DIR_PMASK = $my_pmask_dir;
		     $PMASKFILE = 'I<CH>_pmask.fits'; }
if ($my_pmask) { $PMASKFILE = $my_pmask; }


## ############################################################################
## Read input parameters
## ############################################################################

# Input parameters
my $NARGV = scalar @ARGV;
if ($NARGV<1) { print $usg; exit 1; }
my $proc_dir = shift @ARGV;
if (not $aorlist) { @dirlist = @ARGV; }
else { 
  if ($NARGV>1) {print "$pname: WARNING: using AORs from --aorlist option\n"; } }
my $naors = scalar @dirlist;
if ($naors<1) {
  print "$pname: you should give either a list of AORs or ";
  print "an AOR list file with the --aorlist option\n";
  print $usg; exit 1;
}


## ############################################################################
## Create directory structure and move files
## ############################################################################

# Check if top level directory exists
if (-f $proc_dir) {
  print "$pname: ERROR: cannot create $proc_dir directory. ";
  print "There is already a file with that name\n";
  exit 0;
}
if (-d $proc_dir) {
  print "$pname: ERROR: cannot create $proc_dir directory. ";
  print "The directory already exists\n";
  exit 0;
}

# Create directories
mkdir $proc_dir;
mkdir "$proc_dir/Lists";
mkdir "$proc_dir/reduce";
mkdir "$proc_dir/analysis";

# Copy IRACproc config files
system "cp $ENV{IRAC_PROC}/config/* $proc_dir/Lists/";
system "cp $proc_dir/Lists/my_iracproc.pdl $proc_dir/reduce/";


## ############################################################################
## Set IRACproc and Pmask directories, and Pmask files in configuration files
## ############################################################################

my $ip_path = "$ENV{IRAC_PROC}";

use vars qw($LS);

# Create list
$LS = "ls $proc_dir/Lists/*.par |";
open(LS); my @list = <LS>; close(LS);
chomp @list;

# Set up configuration files
foreach my $parfile (@list) {

    my @config;
    tie(@config,'Tie::File',$parfile) or die "cannot open file $parfile\n";
    for (@config) {
      s/<IRACproc>/$ip_path/g;
      s/<PmaskDir>/$DIR_PMASK/g;
      s/<PmaskFile>/$PMASKFILE/g;
      if ($my_pmask_dir) { s/<IRAC_PROC>\/pmask\///g; }
    }

}


## ############################################################################
## Copy BCD into reduce directory
## ############################################################################

my $func = "cp";
if ($link) { $func = "ln -s"; }
my ($path,$path1,$path2,$path3,$path4);

# Get list of BCD for each AOR in aorlist
use vars qw($LS1 $LS2 $LS3 $LS4);
foreach my $dir (@dirlist) {

    print "$pname: working on AOR $dir which is $proc_dir";

    if (not $bcdpath) {
      # Create list
      $path1 = $dir."/ch1/bcd";
      $path2 = $dir."/ch2/bcd";
      $path3 = $dir."/ch3/bcd";
      $path4 = $dir."/ch4/bcd";
      $LS1 = "ls $path1 |";
      $LS2 = "ls $path2 |";
      $LS3 = "ls $path3 |";
      $LS4 = "ls $path4 |";
      open(LS1); my @list1 = <LS1>; close(LS1);
      open(LS2); my @list2 = <LS2>; close(LS2);
      open(LS3); my @list3 = <LS3>; close(LS3);
      open(LS4); my @list4 = <LS4>; close(LS4);

      # Copy files with new names
      cp_fixup(\@list1,$prefix,$func,$path1,$proc_dir); print " .";
      cp_fixup(\@list2,$prefix,$func,$path2,$proc_dir); print ".";
      cp_fixup(\@list3,$prefix,$func,$path3,$proc_dir); print ".";
      cp_fixup(\@list4,$prefix,$func,$path4,$proc_dir); print ".\n";
      print "$pname has finished copying files for AOR $dir\n";
    }
    else {

      # Create list
      $path = $dir."/".$bcdpath;
      $LS1 = "ls $path |";
      open(LS1); my @list1 = <LS1>; close(LS1);

      # Copy files with new names
      cp_fixup(\@list1,$prefix,$func,$path,$proc_dir); print " ....\n";
      print "$pname has finished copying files for AOR $dir\n";
    }

}


## ############################################################################
## So long and thanks for all the fish
## ############################################################################

exit 0;


## ############################################################################
## Check and copy/link fits file function
## ############################################################################

sub cp_fixup {

    # Read input parameters
    my $list = shift;
    my $prefix = shift;
    my $func = shift;
    my $path = shift;
    my $proc_dir = shift;

    my ($ch,$expid,$dcenum,$filename);

    # Check and copy/link
    foreach (@$list) {
      chomp;
      if (m/^$prefix(I)(\d)_(\d+)_(\d+)_(\d+)_(\d+)_(bcd|bunc)\.fits$/) {
        $ch = $2;
        $expid = sprintf("%04i",$4);
        $dcenum = sprintf("%01i",$5);
        $filename = 'I'.$2.'_'.$3.'_'.$expid.'_'.$dcenum.'_'.$6.'_'.$7.'.fits';


#print "$filename\n";
#print "$func $path/$_ $proc_dir/reduce/$filename\n";
#return;

        system "$func $path/$_ $proc_dir/reduce/$filename";
      }
      if (m/I(\d)_(\d+)_(\d+)_(\d+)_(\d+)_(bdmsk|bimsk)\.fits$/) {
        $ch = $1;
        $expid = sprintf("%04i",$3);
        $dcenum = sprintf("%01i",$4);
        $filename = 'I'.$1.'_'.$2.'_'.$expid.'_'.$dcenum.'_'.$5.'_'.$6.'.fits';
        system "$func $path/$_ $proc_dir/reduce/$filename";
      }
    }

    # Return
    return;

}


1; # Compiler be happy
