## ############################################################################
## refixup - resets configuration files for IRACproc v4.2
##
## (c) 2007 SAO, Mike Schuster - mschuster@cfa.harvard.edu
##
## ############################################################################

=head1 NAME

perl refixup.pl - resets configuration files for IRACproc v4.2


=head1 SYNOPSIS

 % perl refixup.pl [--help][--ver][--pmasks mmmyy][--subarray] proc_dir


=head1 COMMAND LINE OPTIONS

 --help
   Print usage information and exit.

 --ver
   Print current version and exit.

 --pmasks mmmyy
   Set of Pmasks, as distributed by the SSC, specified by month and year
   available options: jun06 (default), dec05, jun05, feb05, jan05, oct04, 
   jun04, mar04

 --subarray
   Set this flag if you are processing data taken in sub-array mode
   not set by default

 proc_dir
   name of the top level directory for IRACproc processing


=head1 DESCRIPTION

This script resets configuration files from an exisiting IRACproc directory 
structure for IRACproc v4.2. This script sets the location of IRACproc and 
other processing parameters in several IRACproc configuration files.


=head1 CHANGES

 1.0 - Initial version (Apr 25, 2007)


=head1 AUTHOR

 (c) 2007 Mike Schuster (mschuster@cfa.harvard.edu)
 (c) 2007 SAO - Smithsonian Astrophysical Observatory


=head1 SEE ALSO

 fixup.pl(1); multi_fixup.pl(1);


=head1 BUGS

Report bugs to Mike Schuster, mschuster@cfa.harvard.edu


=cut

## ############################################################################
## Set Modules and version number
## ############################################################################

use warnings;
use strict;
use diagnostics;
use Getopt::Long;
use Tie::File;

# Set number version
my $VERSION = '1.0';           # Script version number
my $pname   = 'refixup.pl';    # Script name


## ############################################################################
## Read input options
## ############################################################################

my ($help,$ver);
my ($pmasks,$subarray);

my %optctl = (help            => \$help,
              ver             => \$ver,
              pmasks          => \$pmasks,
              subarray        => \$subarray,
             );

&GetOptions(\%optctl,
            "help",
            "ver",
            "pmasks=s",
            "subarray",
           );


## ############################################################################
## Process input options
## ############################################################################

# Print help and exit
my $usg = "USAGE:  perl $pname [--help][--ver][--pmasks mmmyy][--subarray] proc_dir\n";

if ($help) {
  print STDOUT <<HELP;

  $pname version $VERSION

  Reset configuration files for IRACproc v4.2

  $usg

HELP
  exit 0;
}

# Print version and exit
if ($ver) { 
  print "$pname: version $VERSION\n"; 
  exit 0;
}

# Check validity of options
if (not $pmasks) { $pmasks = "jun06"; }
else {
  if (($pmasks ne 'jun06') and ($pmasks ne 'dec05') and ($pmasks ne 'jun05') and ($pmasks ne 'feb05') and ($pmasks ne 'jan05') and ($pmasks ne 'oct04') and ($pmasks ne 'jun04') and ($pmasks ne 'mar04')) {
    print "$pname: ERROR: pmasks must be one of the following:\n";
    print "          jun06, dec05, jun05, feb05, jan05, oct04, jun04, mar04\n";
    exit 0;
  }
}
my $PMASKFILE = "chan<CH>_".$pmasks."_pmask.fits";
if ($subarray) {
  $PMASKFILE = "chan<CH>_subarray_".$pmasks."_pmask.fits"; }
my $DIR_PMASK = "june06_pmasks";
if ($pmasks eq 'dec05') {
  $DIR_PMASK = "december05_pmasks"; }
if ($pmasks eq 'jun05') {
  $DIR_PMASK = "june05_pmasks"; }
if ($pmasks eq 'feb05') {
  $DIR_PMASK = "february05_pmasks"; }
if ($pmasks eq 'jan05') {
  $DIR_PMASK = "january05_pmasks"; }
if ($pmasks eq 'oct04') {
  $DIR_PMASK = "october04_pmasks"; }
if ($pmasks eq 'jun04') {
  $DIR_PMASK = "june04_pmasks"; }
if ($pmasks eq 'mar04') {
  $DIR_PMASK = "march04_pmasks"; }


## ############################################################################
## Read input parameters
## ############################################################################

# Input parameters
my $NARGV = scalar @ARGV;
if ($NARGV<1) { print $usg; exit 1; }
my $proc_dir = shift @ARGV;


## ############################################################################
## Move new configuration files into existing IRACproc directory structure
## ############################################################################

# Check if IRACproc directory structure exists
if (not -d $proc_dir) {
  print "$pname: ERROR: cannot find $proc_dir directory\n";
  exit 0;
}
my $lists_dir = $proc_dir.'/Lists';
if (not -d $lists_dir) {
  print "$pname: ERROR: cannot find $lists_dir directory\n";
  exit 0;
}
my $reduce_dir = $proc_dir.'/reduce';
if (not -d $reduce_dir) {
  print "$pname: ERROR: cannot find $reduce_dir directory\n";
  exit 0;
}

# Copy new IRACproc config files
system "cp -f $ENV{IRAC_PROC}/config/* $proc_dir/Lists/";
system "cp $proc_dir/Lists/my_iracproc.pdl $proc_dir/reduce/";


## ############################################################################
## Set IRACproc and Pmask directories, and Pmask files in configuration files
## ############################################################################

my $ip_path = "$ENV{IRAC_PROC}";

use vars qw($LS);

# Create list
$LS = "ls $proc_dir/Lists/*.par |";
open(LS); my @list = <LS>; close(LS);
chomp @list;

# Reset configuration files
foreach my $parfile (@list) {

    my @config;
    tie(@config,'Tie::File',$parfile) or die "cannot open file $parfile\n";
    for (@config) {
      s/<IRACproc>/$ip_path/g;
      s/<PmaskDir>/$DIR_PMASK/g;
      s/<PmaskFile>/$PMASKFILE/g;
    }

}


## ############################################################################
## Exit
## ############################################################################

exit 0;


1;
